#include "StdAfx.h"
#include "VCamSDKImpl.h"

// DirectShow BaseClasses
#ifdef _DEBUG
#pragma comment(lib, "STRMBASD.LIB")
#else
#pragma comment(lib, "STRMBASE.LIB")
#endif

#define SAFE_RELEASE(x) { if (x) x->Release(); x = NULL; }

// VCam filter interface
#pragma warning(disable : 4192)
#import "VCamFilter.tlb" no_namespace, raw_interfaces_only exclude("UINT_PTR") 

const CLSID CLSID_VCamRenderer = {0x8B9C4F32, 0x044E, 0x491c, {0x89, 0x3E, 0x36, 0x2C, 0xB8, 0xA6, 0x79, 0xD5}};
const IID	IID_IVCamRenderConfig = {0xcc51ba60, 0x7659, 0x4970, {0xb3, 0x21, 0x7d, 0x44, 0x2b, 0xc2, 0x3e, 0x51}};


// Thread used to check filter graph status (whether stopped...)
DWORD WINAPI GraphCheckThread(LPVOID lpParam)
{
	CVCamSDKImpl* pVCam = (CVCamSDKImpl*)lpParam;
	if (pVCam && NULL != pVCam->m_pGraph) {
		IMediaEventEx* pEvent = NULL;
		HRESULT hr = pVCam->m_pGraph->QueryInterface(IID_IMediaEventEx, (void **)&pEvent);

		HANDLE  hEvent; 
		long    evCode, param1, param2;
		BOOLEAN bDone = FALSE;
		hr = pEvent->GetEventHandle((OAEVENT*)&hEvent);
		if (FAILED(hr))
		{
			/* Insert failure-handling code here. */
			SAFE_RELEASE(pEvent);
			return 2;
		}
		while(!bDone && !pVCam->m_bStopGraphCheck && NULL != pVCam->m_pGraph) 
		{
			if (WAIT_OBJECT_0 == WaitForSingleObject(hEvent, 100))
			{ 
				while (hr = pEvent->GetEvent(&evCode, &param1, &param2, 0), SUCCEEDED(hr)) 
				{
					pEvent->FreeEventParams(evCode, param1, param2);
					bDone = (EC_COMPLETE == evCode || EC_USERABORT == evCode);
					ATLTRACE(_T("evCode = 0x%08X\n"), evCode);
				}
			}
		}
		SAFE_RELEASE(pEvent);
		if (bDone) {
			pVCam->StopPlay(); // play to end, ok, let's stop the graph here!
		}
		return 0;
	}
	return 1;
}

DWORD WINAPI ScreenCaptureThread(LPVOID lpParam)
{
	CVCamSDKImpl* pVCam = (CVCamSDKImpl*)lpParam;
	if (pVCam) {
		int x = pVCam->m_rcCaptureZone.left;
		int y = pVCam->m_rcCaptureZone.top;
		int cx = pVCam->m_rcCaptureZone.right - pVCam->m_rcCaptureZone.left;
		int cy = pVCam->m_rcCaptureZone.bottom - pVCam->m_rcCaptureZone.top;

		// Initialize BITMAPINFOHEADER
		BITMAPINFOHEADER bi;
		bi.biSize = sizeof(BITMAPINFOHEADER);
		bi.biWidth = cx;
		bi.biHeight = cy;
		bi.biPlanes = 1;
		bi.biBitCount = 24;
		bi.biCompression = BI_RGB;    
		bi.biSizeImage = ((cx*24+31)/32*4)*cy;
		bi.biXPelsPerMeter = 0;
		bi.biYPelsPerMeter = 0;
		bi.biClrUsed = 0;
		bi.biClrImportant = 0;

		// init GDI objects
		HDC		hdcDesktop	= ::GetDC(HWND_DESKTOP);
		HDC		hdcTemp		= ::CreateCompatibleDC(hdcDesktop);
		LPBYTE	rgb24Buffer	= NULL;
		HBITMAP hbmpTemp	= ::CreateDIBSection(hdcDesktop, (BITMAPINFO*)&bi, DIB_RGB_COLORS, (void**)&rgb24Buffer, 0, 0);
		ATLASSERT(rgb24Buffer);
		HGDIOBJ hgoTemp		= ::SelectObject(hdcTemp, hbmpTemp);

		while (!pVCam->m_bStopCapture) {
			::BitBlt(hdcTemp, 0, 0, cx, cy, hdcDesktop, x, y, SRCCOPY);
			if (pVCam->m_pConfig)
				pVCam->m_pConfig->SendFrameToDriverEx(rgb24Buffer, cx, cy);
			Sleep(40);
		}
		::DeleteObject(::SelectObject(hdcTemp, hgoTemp));
		::DeleteDC(hdcTemp);
		::ReleaseDC(HWND_DESKTOP, hdcDesktop);
	}
	return 0;
}


//////////////////////////////////////////////////////////////////////////


CVCamSDKImpl::CVCamSDKImpl(void)
	: m_pConfig(NULL)
	, m_hGraphCheckThread(NULL)
	, m_bStopGraphCheck(TRUE)
	, m_hScreenCaptureThread(NULL)
	, m_bStopCapture(TRUE)
{
	// ::CoInitialize(NULL);

	HRESULT hr = CoCreateInstance(CLSID_VCamRenderer, NULL, CLSCTX_INPROC, IID_IBaseFilter, 
		reinterpret_cast<void**>(&m_pVCamRenderer));

	if (FAILED(hr)) {
		MessageBox(NULL, _T("Failed to create VCam Renderer filter!"), _T("Error"), MB_OK | MB_ICONERROR);
	}

	// get [IVCamRenderConfig] interface from VCam Renderer filter
	hr = m_pVCamRenderer->QueryInterface(IID_IVCamRenderConfig, reinterpret_cast<void**>(&m_pConfig));
	ATLASSERT(SUCCEEDED(hr));
}

CVCamSDKImpl::~CVCamSDKImpl(void)
{
	StopPlay();
	StopCapture();

	SAFE_RELEASE(m_pConfig);
	// ::CoUninitialize();
}

void	CVCamSDKImpl::SetVCamDeviceName(BSTR pDevName)
{
	if (m_pConfig) {
		m_pConfig->SetVCamDeviceName(pDevName);
	}
}

void	CVCamSDKImpl::SetLicenseCode(BSTR pRegStr)
{
	if (m_pConfig) {
		m_pConfig->SetLicenseCode(pRegStr);
	}
}

//////////////////////////////////////////////////////////////////////////
// DShow Helper functions

// DisconnectFilter - Disconnect all pins on filter.
static void DisconnectFilter(IGraphBuilder *pGraph, IBaseFilter *pFilter)
{
	IEnumPins* pPins;
	IPin* pPin;
	IPin* pConnectedPin;
	DWORD	numPinsFetched;

	HRESULT hr = pFilter->EnumPins(&pPins);
	if (FAILED(hr))
		return;
	pPins->Reset();

	while( SUCCEEDED(pPins->Next(1, &pPin, &numPinsFetched)) && (numPinsFetched == 1) )	{					
		if ( pPin != NULL) {
			pPin->ConnectedTo(&pConnectedPin);
			if (pConnectedPin) {
				pGraph->Disconnect(pPin);
				pGraph->Disconnect(pConnectedPin);
				SAFE_RELEASE(pConnectedPin);
			}
			SAFE_RELEASE(pPin);
		}
	}
	SAFE_RELEASE(pPins);
}

// Disconnect and remove all filters in graph
static HRESULT	TearDownGraph(IGraphBuilder *pGraph)
{
	HRESULT         hr              = S_OK;

	IBaseFilter     *pFilter        = NULL;
	IEnumFilters    *pIFilterEnum   = NULL;


	if (pGraph == NULL) return S_OK;

	// now go unload rendered filters
	hr = pGraph->EnumFilters(&pIFilterEnum);

	if (FAILED(hr)) {
		goto err;
	}

	pIFilterEnum->Reset();

	while ((pIFilterEnum->Next(1, &pFilter, 0) == S_OK) && (pFilter != 0)) {
		DisconnectFilter(pGraph, pFilter);
		if ((hr = pGraph->RemoveFilter(pFilter)) != S_OK) {
			goto err;
		}

		SAFE_RELEASE(pFilter);
		pIFilterEnum->Reset();
	}

err:
	SAFE_RELEASE(pFilter);
	SAFE_RELEASE(pIFilterEnum);

	return hr;
}

//////////////////////////////////////////////////////////////////////////
// play a video file to our virtual camera


struct OverlayConfig {
	int			pos_x, pos_y;
	int			justify_x, justify_y;
	int			opacity;
	bool		bNonPremultAlpha;
	bool		bNeedInvertAlpha;
	int			ol_width, ol_height;
	PVOID		ol_buffer;
};

typedef void (WINAPI *FP_EffectProcess)(PBYTE src, PBYTE dst, DWORD width, DWORD height);
typedef void (WINAPI *FP_GDIDraw)(HDC hdc, DWORD width, DWORD height);

DECLARE_INTERFACE_(IVCamTransConfig, IUnknown)
{
	STDMETHOD(AddImageOverlay) (OverlayConfig* pIOC, int whichOne = 0) PURE;
	STDMETHOD(AddVideoEffect) (LONG effectType, PLONG effectId, LPTSTR sParamFormat) PURE;
	STDMETHOD(SetVideoEffectParams) (LONG effectId, LPCTSTR sParam) PURE;
	STDMETHOD(RemoveVideoEffect) (LONG index = 0) PURE;
	STDMETHOD(RemoveVideoEffects) () PURE;
	STDMETHOD(SetVideoGDIDrawer) (FP_GDIDraw pDrawer) PURE;
	STDMETHOD(SetVideoEffectProcessor) (FP_EffectProcess pProcessor) PURE;
};

void GDIDraw(HDC hdc, DWORD width, DWORD height)
{
	::LineTo(hdc, 100, 100);
}

HRESULT CVCamSDKImpl::PlayVideoFile(LPCTSTR fileName)
{
	StopPlay();
	StopCapture();

	// create filter graph
	HRESULT hr = CoCreateInstance(CLSID_FilterGraph, NULL, 
		CLSCTX_INPROC, IID_IGraphBuilder, (void **)&m_pGraph);
	
	hr = m_pGraph->AddFilter(m_pVCamRenderer, L"Renderer");

#if 0 // Use "e2eSoft Effects Filter" - TEST
	const CLSID CLSID_VCamTrans = {0x9CFB3966,0xB84C,0x4452,{0x9A,0x62,0x33,0xA2,0xC6,0x6E,0xEE,0xB1}};
	const IID	IID_IVCamTransConfig = {0x799b6eca, 0xbc9f, 0x4bd2, {0x91, 0x80, 0x45, 0x68, 0x46, 0x10, 0x42, 0xcd}};

	IBaseFilter* pVCamTrans = NULL;
	hr = CoCreateInstance(CLSID_VCamTrans, NULL, CLSCTX_INPROC, IID_IBaseFilter, (void **)&pVCamTrans);
	hr = m_pGraph->AddFilter(pVCamTrans, L"VCam Effects Filter");
	IVCamTransConfig* pConfig;
	hr = pVCamTrans->QueryInterface(IID_IVCamTransConfig, reinterpret_cast< void** >(&pConfig));
	LONG id;
	TCHAR szParams[] = _T("bright[-255..255]={100}, contrast[0..32]={10}");
	hr = pConfig->AddVideoEffect(0, &id, szParams);
	hr = pConfig->SetVideoGDIDrawer(&GDIDraw);
#endif

	hr = m_pGraph->RenderFile(CT2W(fileName), NULL);
	if (FAILED(hr)) 
		return hr;

	m_pMC = m_pGraph;
	if (NULL != m_pMC) m_pMC->Run();
	m_pMS = m_pGraph;

	// -
	m_bStopGraphCheck = FALSE;
	DWORD threadId;
	m_hGraphCheckThread = ::CreateThread(NULL, 0, GraphCheckThread, this, 0, &threadId);
	return S_OK;
}


LONG	CVCamSDKImpl::GetCurrentVideoDuration()
{
	if ((m_pGraph != NULL) && (m_pMS != NULL)) {
		REFERENCE_TIME rtTotalTime = 0;
		if (SUCCEEDED(m_pMS->GetDuration(&rtTotalTime)))
			return (LONG)(rtTotalTime / (UNITS / MILLISECONDS));
	}

	return 0L;
}

LONG	CVCamSDKImpl::GetCurrentVideoPosition()
{
	if ((m_pGraph != NULL) && (m_pMS != NULL)) {
		REFERENCE_TIME timeNow;
		if (SUCCEEDED(m_pMS->GetCurrentPosition(&timeNow)))
			return (LONG)(timeNow / (UNITS / MILLISECONDS));
	}

	return 0L;
}

HRESULT	CVCamSDKImpl::SetCurrentVideoPosition(LONG pos)
{
	if ((m_pGraph != NULL) && (m_pMS != NULL)) {
		REFERENCE_TIME rtPos = pos * (UNITS / MILLISECONDS);
		return m_pMS->SetPositions(&rtPos, AM_SEEKING_AbsolutePositioning, 
			NULL, AM_SEEKING_NoPositioning);
	}

	return E_FAIL;
}


bool CVCamSDKImpl::IsPlaying() 
{ 
	if ((m_pGraph != NULL) && (m_pMC != NULL)) {
		FILTER_STATE fs;
		m_pMC->GetState(100, (OAFilterState *)&fs);
		return (State_Running == fs); 
	}

	return false;
}

//////////////////////////////////////////////////////////////////////////
// play another webcam's video to our virtual camera
HRESULT CVCamSDKImpl::PlayVideoDevice(UINT idx)
{
	StopPlay();
	StopCapture();

	// enumerate all video capture devices
	ICreateDevEnum*	pCreateDevEnum;
	IEnumMoniker*	pEnum;	
	IMoniker*		pM = NULL;
	ULONG			cFetched;

	// let's get a device enumerator.
	HRESULT hr = CoCreateInstance(CLSID_SystemDeviceEnum, NULL, 
		CLSCTX_INPROC_SERVER, IID_ICreateDevEnum, (void **)&pCreateDevEnum);
	if (FAILED(hr)) return hr;

	// get an enumerator to enumerate video capture devices.
	hr = pCreateDevEnum->CreateClassEnumerator(CLSID_VideoInputDeviceCategory, &pEnum, 0);
	if (FAILED(hr)) return hr;

	// reset the enumerators position.
	if (pEnum) {
		hr = pEnum->Reset();

		// while we have devices to grab let's get them.
		int i=0;
		while (pEnum->Next(1, &pM, &cFetched) == S_OK) {			
			IPropertyBagPtr pBag;
			HRESULT hResult = pM->BindToStorage(0, 0, IID_IPropertyBag, (void **)&pBag);
			if (idx == i++)
				break;
			SAFE_RELEASE(pM);
		}
		SAFE_RELEASE(pEnum);
	}
	SAFE_RELEASE(pCreateDevEnum);

	// we found it
	if (pM) {
		ICaptureGraphBuilder2*	pBuilder = NULL;
		IBaseFilter*	pCapSrcFilter = NULL;
		IBindCtx *pbc = NULL; 
		CreateBindCtx(0, &pbc); 
		hr = pM->BindToObject(pbc, 0, IID_IBaseFilter, (void**)&pCapSrcFilter);
		SAFE_RELEASE(pbc);

		// now create the filter graph
		hr = CoCreateInstance(CLSID_CaptureGraphBuilder2, NULL, 
			CLSCTX_INPROC_SERVER, IID_ICaptureGraphBuilder2, (void**)&pBuilder);

		hr = CoCreateInstance(CLSID_FilterGraph, NULL, 
			CLSCTX_INPROC, IID_IGraphBuilder, (void **)&m_pGraph);

		pBuilder->SetFiltergraph(m_pGraph);

		hr = m_pGraph->AddFilter(pCapSrcFilter, L"Source");
		hr = m_pGraph->AddFilter(m_pVCamRenderer, L"Renderer");
		hr = pBuilder->RenderStream(NULL, NULL, pCapSrcFilter, NULL, m_pVCamRenderer);
		if (FAILED(hr)) return hr;

		m_pMC = m_pGraph;
		if (NULL != m_pMC) m_pMC->Run();
		m_pMS = m_pGraph;

		SAFE_RELEASE(pCapSrcFilter);
		SAFE_RELEASE(pM);
		SAFE_RELEASE(pBuilder);
	}
	return S_OK;
}

// stop video file or device play
void	CVCamSDKImpl::StopPlay()
{
	if (NULL != m_pGraph) {
		m_pMC->Stop();

		// -
		if (m_hGraphCheckThread) {
			m_bStopGraphCheck = TRUE;
			if (WAIT_TIMEOUT == WaitForSingleObject(m_hGraphCheckThread, 100))
				TerminateThread(m_hGraphCheckThread, 0);
			CloseHandle(m_hGraphCheckThread);
			m_hGraphCheckThread = NULL;
		}

		TearDownGraph(m_pGraph);
		m_pGraph = NULL;
		m_pMC = NULL;
		m_pMS = NULL;
	}
}

void	CVCamSDKImpl::CaptureScreen(RECT rcZone)
{
	StopPlay();
	StopCapture();

	m_rcCaptureZone = rcZone;
	m_bStopCapture = FALSE;
	DWORD threadId;
	m_hScreenCaptureThread = ::CreateThread(NULL, 0, ScreenCaptureThread, this, 0, &threadId);
}

void	CVCamSDKImpl::StopCapture()
{
	if (m_hScreenCaptureThread) {
		m_bStopCapture = TRUE;
		if (WAIT_TIMEOUT == WaitForSingleObject(m_hScreenCaptureThread, 100))
			TerminateThread(m_hScreenCaptureThread, 0);
		CloseHandle(m_hScreenCaptureThread);
		m_hScreenCaptureThread = NULL;
	}
}

// set RGB24 image buffer as current virtual camera output
void	CVCamSDKImpl::PlayBuffer(PBYTE rgb24Buffer, DWORD dwWidth, DWORD dwHeight)
{
	if (m_pConfig) {
		m_pConfig->SendFrameToDriverEx(rgb24Buffer, dwWidth, dwHeight);
	}
}

void	CVCamSDKImpl::AddOverlay(BYTE *pRgb32Buffer, long width, long height, long px, long py, long whichOne)
{
	if (m_pConfig) {
		m_pConfig->AddOverlay(pRgb32Buffer, width, height, px, py, whichOne);
	}
}

void	CVCamSDKImpl::AddText(LPCTSTR olText, PLOGFONT pLogFont, COLORREF ftColor, COLORREF bkColor, long px, long py, long bkType, long whichOne)
{
	if (m_pConfig) {
		m_pConfig->AddText(_bstr_t(olText), (BYTE*)pLogFont, ftColor, bkColor, px, py, bkType, whichOne);
	}
}

// change driver configuration demo : change default frame rate
void	CVCamSDKImpl::ChangeDefaultFPS(int fps)
{
	if (m_pConfig) {
		m_pConfig->SetDefaultFPS(fps);
	}
}

void	CVCamSDKImpl::ChangeDefaultVideoSize(int szType)
{
	if (m_pConfig) {
		m_pConfig->SetDefaultVCamSize(szType);
	}
}

bool CVCamSDKImpl::GetOnlySupportOneSize()
{
	long onesize = 0;
	m_pConfig->GetOnlyOneSize(&onesize);
	return onesize != 0;
}

void CVCamSDKImpl::SetOnlySupportOneSize(bool onesize)
{
	if (m_pConfig) {
		m_pConfig->SetOnlyOneSize(onesize);
	}
}

void CVCamSDKImpl::SetCustomSize(long use_custom_size, long width, long height)
{
	if (m_pConfig) {
		m_pConfig->SetCustomSize(use_custom_size, width, height);
	}
}
